#!/usr/bin/python

import BluebagConfig

import logging, sys, time, traceback
from os import path
import MySQLdb as mysql_api
#from pysqlite2 import dbapi2 as sqlite_api

class DatabaseManager:
  
  def __init__(self, bluebagConfig):
    self.logger_ = logging.getLogger('DatabaseManager');
    self.bluebagConfig_ = bluebagConfig;    
    self.dbConnection_ = None;
    self.dbCursor_ = None;
    self.initialized_ = False;
    self.initialized_ = self.initDb();
    if self.initialized_: self.logger_.info("DatabaseManager initialized.");
    else: self.logger_.fatal("Error while initializing the database.");

#  def busy_handler(self, a, b, c):
#    self.logger_.debug("Busy waiting for database to unlock!");
#    sleep(0.5);
#    return 1;
    
  def connect(self):
    if not self.initialized_:
      self.logger_.error("Database not initialized.");
      return False;
    if self.dbConnection_:
      self.logger_.error("Already connected.");
      return False;

    self.logger_.debug("Connecting to the database.");
    if self.bluebagConfig_.dbType_ == "mysql":
      self.dbConnection_ = mysql_api.connect(unix_socket=self.bluebagConfig_.dbSocket_, db=self.bluebagConfig_.dbName_, user=self.bluebagConfig_.dbUser_, passwd=self.bluebagConfig_.dbPassword_);
    else:
      self.dbConnection_ = sqlite_api.connect(self.bluebagConfig_.dbDir_+self.bluebagConfig_.dbName_+".db",timeout=20.0);
    #self.dbCursor_ = self.dbConnection_.cursor();

    return True;
  
  def disconnect(self):
    if not self.dbConnection_: # or not self.dbCursor_:
      self.logger_.error("Not connected.");
      return False;
    self.logger_.debug("Disconnecting from the database.");
    self.dbConnection_.commit();
    if self.dbCursor_: self.dbCursor_.close();
    self.dbConnection_.close();
    self.dbConnection_ = None;
    self.dbCursor_ = None;
    return True;
  
  def execute(self, statement):
    if not self.initialized_:
      self.logger_.error("Database not initialized.");
      return None;
    if not self.dbCursor_:
      self.dbCursor_ = self.dbConnection_.cursor();
    #  self.logger_.error("Not connected.");
    #  return False;
    for count in range(10):
      try:
        self.logger_.debug("Executing statement: "+statement);
        self.dbCursor_.execute(statement);
        break;
      except:
        self.logger_.fatal("?!??!: " + traceback.format_exc());
      self.logger_.warning("Sleeping before retrying to execute.");
      time.sleep(1);
    return self.dbCursor_;

  def closeCursor(self):
    if self.dbCursor_:
      self.dbCursor_.close();
      self.dbCursor_ = None;

  def commit(self):
    if not self.initialized_:
      self.logger_.error("Database not initialized.");
      return False;
    if not self.dbConnection_:
      self.logger_.error("Not connected.");
      return False;
    res = False;
    for count in range(10):
      try:
        self.logger_.debug("Committing transaction.");
        self.dbConnection_.commit();
        if self.dbCursor_:
          self.dbCursor_.close();
          self.dbCursor_ = None;
        self.logger_.debug("Transaction committed.");
        res = True;
        break;
      except:
        self.logger_.fatal("?!??!: " + traceback.format_exc());
        res = False;
      self.logger_.warning("Sleeping before retrying to commit.");
      time.sleep(1);
    return res;

  def initDb(self):
    try:
        self.logger_.info("Initializing databases.");

        if self.bluebagConfig_.dbType_ == "mysql":
          self.dbConnection_ = mysql_api.connect(unix_socket=self.bluebagConfig_.dbSocket_, db=self.bluebagConfig_.dbName_, user=self.bluebagConfig_.dbUser_, passwd=self.bluebagConfig_.dbPassword_);
        else:
          self.dbConnection_ = sqlite_api.connect(self.dbPath_+self.dbName_+".db",timeout=20.0);
        self.dbCursor_ = self.dbConnection_.cursor();
        
        if self.bluebagConfig_.dbType_ == "mysql":
          self.dbCursor_.execute("show tables");
        else:  
          self.dbCursor_.execute("select name from sqlite_master where type='table'");
        res = self.dbCursor_.fetchall();
        table_list = [];
        for result in res:
          table_list.append(result[0]);
        if "btdevice" not in table_list:
          self.logger_.debug("Connecting to the database.");
          self.logger_.debug("Creating btdevice table");
          if self.bluebagConfig_.dbType_ == "mysql":
            self.dbCursor_.execute("CREATE TABLE btdevice (" + \
                              "bt_address character(20)," + \
                              "bt_name character varying(200)," + \
                              "man_id character varying(200)," + \
                              "bt_signal integer," + \
                              "bt_firstseen timestamp not null default '0000-00-00 00:00:00'," + \
                              "bt_lastseen timestamp not null default '0000-00-00 00:00:00'," + \
                              "bt_info character varying(1000)," + \
                              "bt_servicelist character varying(5000)," + \
                              "bt_timesseen integer DEFAULT 0," + \
                              "bt_tag character varying(500)," + \
                              "bt_task character varying(20)," + \
                              "stupid boolean default false," + \
                              "stupidRetry integer DEFAULT 0);");
          else:
            self.dbCursor_.execute("CREATE TABLE btdevice (" + \
                              "bt_address character(20)," + \
                              "bt_name character varying(200)," + \
                              "man_id character varying(200)," + \
                              "bt_signal integer," + \
                              "bt_firstseen timestamp without time zone," + \
                              "bt_lastseen timestamp without time zone," + \
                              "bt_info character varying(1000)," + \
                              "bt_servicelist character varying(5000)," + \
                              "bt_timesseen integer DEFAULT 0," + \
                              "bt_tag character varying(500)," + \
                              "bt_task character varying(20)," + \
                              "stupid boolean default false," + \
                              "stupidRetry integer DEFAULT 0);");

        self.dbConnection_.commit();
        self.dbCursor_.close();
        self.dbConnection_.close();
        self.dbConnection_ = None;
        self.dbCursor_ = None;
    except:
      self.logger_.fatal("Error while initializing the DB: " + traceback.format_exc());
      return False;
    else:
      self.logger_.info("DB initialized.");
      return True;
      
