#!/usr/bin/python

import BluebagConfig
from QueueManager import QueueManager;
from HciWorker import HciWorker
from DongleManager import DongleManager
#from DatabaseManager import DatabaseManager
import logging, sys, os, time, subprocess, signal, traceback

class FangWorker(HciWorker):

  def __init__(self, dongleManager, queueManager, bluebagConfig):
    HciWorker.__init__(self, dongleManager, bluebagConfig);
    self.queueManager_ = queueManager;
    self.forcedDongles_ = [];
    self.numUsedDongles_ = 1;
    self.workClasses_ = ["fang"];
    self.fangProcess_ = None;
    self.remainingRange_ = "";
    self.logger_.info("FangeWorker initialized.");

  def run(self):
    HciWorker.run(self);
  
  def doProcessing(self):
    # Do the actual work using the assigned dongles
    stdin = None;
    stdout = None;
    stderr = None;
    try:
      self.logger_.debug("Fanging ...");
      
      dongle_ids = "";
      for dongle in self.dongles_:
        if dongle_ids != "": dongle_ids += ",";
        dongle_ids += dongle[3:];
      range = self.device_.replace(":","");
      
      self.logger_.debug("Executing: "+self.bluebagConfig_.redPath_+" -e "+dongle_ids+" -r "+range);
      self.fangProcess_ = subprocess.Popen(self.bluebagConfig_.redPath_+" -e "+dongle_ids+" -r "+range, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True);
      stdin, stdout, stderr = self.fangProcess_.stdin, self.fangProcess_.stdout, self.fangProcess_.stderr;

      output = "";
      self.logger_.debug("Parsing fang output.");
      
      while (self.fangProcess_ != None) and (self.fangProcess_.poll() == None):
        for char in stdout.readline():
          if char == "\n":
            #self.logger_.debug("Fang output: "+output);
            if output[0:6] == "Found:":
              hidden = [output.strip()[-18:-1]];
              self.logger_.info("Found hidden device: "+hidden[0]);
              if self.queueManager_: self.queueManager_.signalNewDevices(hidden);
            elif output[0:4] == "Done":
              self.remainingRange_ = output.strip()[-17:]+self.device_[-18:];
              self.logger_.debug("Remaining range: "+self.remainingRange_);
            output = "";
          else:
            output += char;

      self.logger_.debug("Fang has quit.");
      
    except:
      exctype, excvalue = sys.exc_info()[:2];
      self.logger_.warning("Caught exception in fang worker, trying to ignore it: "+str(excvalue));
      self.cleanup();

    if stdin: stdin.close();
    stdin = None;
    if stdout: stdout.close();
    stdout = None;
    if stderr: stderr.close();
    stderr = None;

    for dongle in self.dongles_:
      self.logger_.debug("Executing: hciconfig "+dongle+" reset");
      os.system("hciconfig "+dongle+" reset");
    # Give some time to dongles to reset properly
    time.sleep(0.3);

#  def signalQueueListener(self, device, work, update):
#    self.updateDevice_ = update;
#    HciWorker.signalQueueListener(self,device,work);

  def stopFang(self):
    self.logger_.debug("Forcing fang to stop immediately.");
    if self.fangProcess_:
      try:
        if self.fangProcess_.poll() == None:
          os.kill(self.fangProcess_.pid, signal.SIGKILL);
          os.waitpid(self.fangProcess_.pid,0);
          #self.fangProcess_.wait();
      except:
        self.logger_.debug("Exception killing fang: "+traceback.format_exc());
	self.fangProcess_ = None;
    
  def quitASAP(self):
    self.logger_.debug("Request the worker to quit ASAP.");
    # Kill external processes and perform specific cleanup
    # in order to be able to quit ASAP
    self.stopFang();
    HciWorker.quitASAP(self);

  def cleanup(self):
    # Insert here cleanup to be done right before stopping the thread
#   self.stopFang();
    HciWorker.cleanup(self);
  
def main():
  # Setting up logging facilities
  logging.basicConfig(level=logging.DEBUG,
                  format='%(asctime)s %(module)s %(levelname)s %(message)s',
                  filename="./fangworker.log",
                  filemode='w');
  console = logging.StreamHandler();
  console.setLevel(logging.DEBUG);
  formatter = logging.Formatter("%(asctime)s %(module)s %(levelname)s %(message)s");
  console.setFormatter(formatter);
  logging.getLogger('').addHandler(console);
  
  # Do some testing
  dongleManager = DongleManager();
  fangWorker = FangWorker(dongleManager, None);
  try:
    fangWorker.start();
    fangWorker.signalQueueListener("00:00:00:00:00:00-00:00:00:00:00:11", ["fang"]);
    time.sleep(3);
    fangWorker.stopFang();
    time.sleep(3);
    fangWorker.signalQueueListener("00:00:00:00:00:00-00:00:00:00:00:11", ["fang"]);
    while fangWorker.busy_: time.sleep(1);
    fangWorker.quitASAP();
    fangWorker.join();
  except:
    exctype, excvalue = sys.exc_info()[:2];
    logging.getLogger('').fatal("?!??!: " + str(excvalue));
    fangWorker.quitASAP();
    fangWorker.join();

if __name__ == "__main__":
  main();
