#!/usr/bin/python

import BluebagConfig
from QueueManager import QueueManager;
from HciWorker import HciWorker
from DongleManager import DongleManager
from DatabaseManager import DatabaseManager
import _mysql
import logging, sys, os, time, traceback

class HciDeviceWorker(HciWorker):

  def __init__(self, dongleManager, queueManager, bluebagConfig, taskTag):
    HciWorker.__init__(self, dongleManager, bluebagConfig);
    self.queueManager_ = queueManager;
    self.taskTag_ = taskTag;
    self.updateDevice_ = False;
    self.forcedDongles_ = [];
    self.numUsedDongles_ = 1;
    self.workClasses_ = ["name", "lq", "info", "sdp", "stupid"];
    self.logger_.info("HciDeviceWorker initialized.");

  def run(self):
    self.databaseManager_ = DatabaseManager(self.bluebagConfig_);
    self.databaseManager_.connect();
    HciWorker.run(self);
  
  def hcitoolConnect(self, dongle, device):
    stdin = None;
    stdouterr = None;
    self.logger_.debug("Executing: "+self.bluebagConfig_.hcitoolPath_+" -i "+dongle+" cc "+device);
    stdin, stdouterr = os.popen4(self.bluebagConfig_.hcitoolPath_+" -i "+dongle+" cc "+device);
    output = str.strip(stdouterr.readline());
    if stdin: stdin.close();
    if stdouterr: stdouterr.close();
    if output != "":
      self.logger_.warning("Error while connecting to "+device+" : "+output);
      return False;
    return True;

  def hcitoolDisconnect(self, dongle, device):
    stdin = None;
    stdouterr = None;
    self.logger_.debug("Executing: "+self.bluebagConfig_.hcitoolPath_+" -i "+dongle+" dc "+device);
    stdin, stdouterr = os.popen4(self.bluebagConfig_.hcitoolPath_+" -i "+dongle+" dc "+device);
    output = str.strip(stdouterr.readline());
    if stdin: stdin.close();
    if stdouterr: stdouterr.close();
    if output != "":
      self.logger_.warning("Error while disconnecting to "+device+" : "+output);
      return False;
    return True;
  
  def doProcessing(self):
    # Do the actual work using the assigned dongle
    stdin = None;
    stdout = None;
    stderr = None;
    stdouterr = None;
    try:
      self.logger_.debug("Processing...");
      if not self.simulation_:
        # name
        name = "";
        if "name" in self.workInProgress_:
          self.logger_.debug("Executing: "+self.bluebagConfig_.hcitoolPath_+" -i "+self.dongles_[0]+" name "+self.device_);
          stdin, stdouterr = os.popen4(self.bluebagConfig_.hcitoolPath_+" -i "+self.dongles_[0]+" name "+self.device_);
          name = str.strip(stdouterr.readline());
          if stdin: stdin.close();
          stdin = None;
          if stdouterr: stdouterr.close();
          stdouterr = None;
          if name == "": self.logger_.warning("Error while retrieving name from device: "+self.device_);
          else: self.logger_.debug("Executed \"name\" command. Result: "+name);
    
        # info
        info = "";
        err = "";
        if "info" in self.workInProgress_:
          self.logger_.debug("Executing: "+self.bluebagConfig_.hcitoolPath_+" -i "+self.dongles_[0]+" info "+self.device_);
          stdin, stdout, stderr = os.popen3(self.bluebagConfig_.hcitoolPath_+" -i "+self.dongles_[0]+" info "+self.device_);
          stdout.readline();
          for line in stdout.readlines():
            info = info + str.strip(line) + "\n";
          info = str.strip(info);
          for line in stderr.readlines():
            err = err + str.strip(line) + "\n";
          err = str.strip(err);
          if stderr: stderr.close();
          stderr = None;
          if stdout: stdout.close();
          stdout = None;
          if stdin: stdin.close();
          stdin = None;
          if err != "":
            self.logger_.warning("Error while retrieving info from device: "+self.device_+" : "+err);
            info = "";
          else: self.logger_.debug("Executed \"info\" command. Result: "+info);
    
        # lq
        connected = False;
        lq = "";
        err = "";
        if "lq" in self.workInProgress_:
          if not connected:
            connected = self.hcitoolConnect(self.dongles_[0],self.device_);
          if connected:
            self.logger_.debug("Executing: "+self.bluebagConfig_.hcitoolPath_+" -i "+self.dongles_[0]+" lq "+self.device_);
            stdin, stdout, stderr = os.popen3(self.bluebagConfig_.hcitoolPath_+" -i "+self.dongles_[0]+" lq "+self.device_);
            for line in stdout.readlines():
              lq = lq + str.strip(line) + "\n";
            lq = str.strip(lq);
            for line in stderr.readlines():
              err = err + str.strip(line) + "\n";
            err = str.strip(err);
            if stderr: stderr.close();
            stderr = None;
            if stdout: stdout.close();
            stdout = None;
            if stdin: stdin.close();
            stdin = None;
            if err != "":
              self.logger_.warning("Error while retrieving quality from device: "+self.device_+" : "+err);
              lq = "-1";
            else:
              self.logger_.debug("Executed \"lq\" command. Result: "+lq);
              lq = lq[lq.find(":")+2:];
            self.logger_.debug("Current link quality for device "+self.device_+": "+lq);
          else:
            self.logger_.warning("Error while retrieving quality from device: "+self.device_+" : "+err);
            lq = "-1";
        if connected:
          self.hcitoolDisconnect(self.dongles_[0],self.device_);
          connected = False;
        if lq == "":
          lq = "-1";
        
        # sdp
        servicelist = "";
        err = "";
        if "sdp" in self.workInProgress_:
          self.logger_.debug("Executing: "+self.bluebagConfig_.sdptoolPath_+" -i "+self.dongles_[0]+" browse "+self.device_);
          stdout = os.popen(self.bluebagConfig_.sdptoolPath_+" -i "+self.dongles_[0]+" browse "+self.device_);
          stdout.readline();
          for line in stdout.readlines():
            servicelist = servicelist + str.strip(line) + "\n";
          servicelist = str.strip(servicelist);
          exitCode = stdout.close();
          stdout = None;
          if exitCode:
            self.logger_.warning("Error while browsing service list from device: "+self.device_);
            servicelist = "";
          else: self.logger_.debug("Executed sdp \"browse\" command. Result: "+servicelist);

        # stupid
        stupid = "";
        btobex_output = "";
        err = "";
        if "stupid" in self.workInProgress_:
          self.logger_.debug("Executing: "+self.bluebagConfig_.btobexPath_+" -i "+self.dongles_[0]+" push "+self.device_+" mik.jpg");
          stdin, stdout, stderr = os.popen3(self.bluebagConfig_.btobexPath_+" -i "+self.dongles_[0]+" push "+self.device_+" mik.jpg");
          for line in stdout.readlines():
            btobex_output = btobex_output + str.strip(line) + "\n";
          btobex_output = str.strip(btobex_output);
          for line in stderr.readlines():
            err = err + str.strip(line) + "\n";
          err = str.strip(err);
          if stderr: stderr.close();
          stderr = None;
          if stdout: stdout.close();
          stdout = None;
          if stdin: stdin.close();
          stdin = None;
          if (btobex_output != "") or (err != ""):
            self.logger_.warning("Error while pushing obex object: "+self.device_);
            stupid = "0";
          else:
            self.logger_.debug("Executed obex push. Result: "+btobex_output);
            stupid = "1";
      else:
        time.sleep(2);
        name = "simulation";
        lq = "128";
        info = "simulation";
        servicelist = "simulation";
        stupid = "0";
      
      # timestamp
      if self.bluebagConfig_.dbType_ == "mysql":
        ts_first = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime());
        ts_last = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime());
      else:
        ts_first = str(time.time());
        ts_last = str(time.time());
      
      # times seen
      timesseen = 1;
      stupidRetry = 1;
      cursor = self.databaseManager_.execute("SELECT bt_timesseen, bt_tag, stupidRetry FROM btdevice WHERE bt_address = '"+self.device_+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
      res = cursor.fetchall();
      self.databaseManager_.closeCursor();
      if len(res) > 0:
        timesseen = res[0][0] + 1;
        stupidRetry = res[0][2] + 1;
      # check the survey tag
      tag = self.taskTag_;
      if (len(res) > 0) and (len(res[0][1]) > 0):
        currentTags = res[0][1].split(',');
        tagFound = False;
        for currentTag in currentTags:
          if currentTag == self.taskTag_: tagFound = True;
        if not tagFound: tag = tag+","+res[0][1];
        else: tag = "";
  
      # man_id
      if not self.updateDevice_:
        deviceManId = self.device_[:8];
        deviceManDesc = "unknown";
        cursor = self.databaseManager_.execute("SELECT man_desc FROM manufacturer WHERE man_id = '"+deviceManId+"'");
        res = cursor.fetchall();
        self.databaseManager_.closeCursor();
        if len(res) > 0:
          deviceManDesc = res[0][0];
        if self.bluebagConfig_.dbType_ == "mysql":
          deviceManDesc = _mysql.escape_string(deviceManDesc);
        else:
          deviceManDesc = deviceManDesc.replace("'","''");
    
      # do some little escaping
      if self.bluebagConfig_.dbType_ == "mysql":
        name = _mysql.escape_string(name);
      else:
        name = name.replace("'","''");
      if self.bluebagConfig_.dbType_ == "mysql":
        info = _mysql.escape_string(info);
      else:
        info = info.replace("'","''");
      if self.bluebagConfig_.dbType_ == "mysql":
        servicelist = _mysql.escape_string(servicelist);
      else:
        servicelist = servicelist.replace("'","''");
      if self.bluebagConfig_.dbType_ == "mysql":
        tag = _mysql.escape_string(tag);
      else:
        tag = tag.replace("'","''");
      # insert
      if self.updateDevice_:
        if name != "":
          self.logger_.info("Updating name for device: "+self.device_);
          self.databaseManager_.execute("UPDATE btdevice SET bt_name = '"+name+"' where bt_address = '"+self.device_+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
        if (lq != "-1") and (lq != ""):
          self.logger_.info("Updating lq for device: "+self.device_);
          self.databaseManager_.execute("UPDATE btdevice SET bt_signal = "+lq+" where bt_address = '"+self.device_+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
        if info != "":
          self.logger_.info("Updating info for device: "+self.device_);
          self.databaseManager_.execute("UPDATE btdevice SET bt_info = '"+info+"' where bt_address = '"+self.device_+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
        if servicelist != "":
          self.logger_.info("Updating servicelist for device: "+self.device_);
          self.databaseManager_.execute("UPDATE btdevice SET bt_servicelist = '"+servicelist+"' where bt_address = '"+self.device_+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
        if tag != "":
          self.logger_.info("Updating tag for device: "+self.device_);
          self.databaseManager_.execute("UPDATE btdevice SET bt_tag = '"+tag+"' where bt_address = '"+self.device_+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
        if stupid != "":
          self.logger_.info("Updating stupid for device: "+self.device_);
          self.databaseManager_.execute("UPDATE btdevice SET stupid = "+str(stupid)+", stupidRetry = "+str(stupidRetry)+" where bt_address = '"+self.device_+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
        self.logger_.info("Updating lastseen and timesseen from device: "+self.device_);
        self.databaseManager_.execute("UPDATE btdevice SET bt_lastseen = '"+ts_last+"', bt_timesseen = "+str(timesseen)+" where bt_address = '"+self.device_+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
      else:
        if stupid == "":
          stupid = "0";
          stupidRetry = 0;
        self.logger_.info("Inserting new device in the database: "+self.device_);
        self.databaseManager_.execute("INSERT INTO btdevice VALUES ("+\
                                       "'"+self.device_+"',"+\
                                       "'"+name+"',"+\
                                       "'"+deviceManDesc+"',"+\
                                       str(lq)+",'"+\
                                       ts_first+"','"+\
                                       ts_last+"',"+\
                                       "'"+info+"',"+\
                                       "'"+servicelist+"',"+\
                                       str(timesseen)+","+\
                                       "'"+tag+"',"+\
                                       "'"+self.bluebagConfig_.task_+"',"+\
                                       str(stupid)+","+\
                                       str(stupidRetry)+")");
      if not self.databaseManager_.commit():
        self.logger_.error("Failed to commit transaction to the database!");
      
      # Signal task completion to the queue manager
      self.queueManager_.signalCompletion(self.device_, self.workInProgress_);

    except:
      self.logger_.warning("Caught exception in device worker, trying to ignore it: "+traceback.format_exc());
      self.queueManager_.signalCompletion(self.device_, self.workInProgress_);
      
    if stdin: stdin.close();
    stdin = None;
    if stdout: stdout.close();
    stdout = None;
    if stderr: stderr.close();
    stderr = None;
    if stdouterr: stdouterr.close();
    stdouterr = None;
    

  def signalQueueListener(self, device, work, update):
    self.updateDevice_ = update;
    print "*********";
    print self.queueManager_.simulationList_;
    print "*********";
    if device in self.queueManager_.simulationList_:
      self.logger_.debug("###### SIMULATION ######");
      self.simulation_ = True;
    else: self.simulation_ = False;
    HciWorker.signalQueueListener(self, device, work);
    
  def quitASAP(self):
    self.logger_.debug("Request the worker to quit ASAP.");
    # Kill external processes and perform specific cleanup
    # in order to be able to quit ASAP
    HciWorker.quitASAP(self);

  def cleanup(self):
    # Insert here cleanup to be done right before stopping the thread
    self.databaseManager_.commit();
    self.databaseManager_.disconnect();
    HciWorker.cleanup(self);
  
def main():
  # Setting up logging facilities
  logging.basicConfig(level=logging.DEBUG,
                  format='%(asctime)s %(module)s %(levelname)s %(message)s',
                  filename="./hcideviceworker.log",
                  filemode='w');
  console = logging.StreamHandler();
  console.setLevel(logging.DEBUG);
  formatter = logging.Formatter("%(asctime)s %(module)s %(levelname)s %(message)s");
  console.setFormatter(formatter);
  logging.getLogger('').addHandler(console);
  
  # Do some testing
  dongleManager = DongleManager();
  dongleManager.addDongle("hci0");
  hciDeviceWorker = HciDeviceWorker(dongleManager, None);
  try:
    hciDeviceWorker.start();
    hciDeviceWorker.signalQueueListener("00:00:00:00:00:00", ["lq"], True);
    while hciDeviceWorker.busy_: time.sleep(1);
    hciDeviceWorker.quitASAP();
    hciDeviceWorker.join();
  except:
    exctype, excvalue = sys.exc_info()[:2];
    logging.getLogger('').fatal("?!??!: " + str(excvalue));
    hciDeviceWorker.quitASAP();
    hciDeviceWorker.join();

if __name__ == "__main__":
  main();
