#!/usr/bin/python

from DongleManager import DongleManager

import logging, sys, time
from random import Random
from threading import Event, Thread

class HciWorker(Thread):
  
  def __init__(self, dongleManager, bluebagConfig):
    Thread.__init__(self);
    
    self.logger_ = logging.getLogger('HciWorker');
    self.quitEvent_ = Event();
    self.wakeupEvent_ = Event();
    self.device_ = "";
    self.dongles_ = [];
    self.busy_ = False;
    self.simulation_ = False;
    self.workInProgress_ = [];
    self.dongleManager_ = dongleManager;
    self.bluebagConfig_ = bluebagConfig;    
    
    self.randomSource_ = Random();
    self.randomSource_.jumpahead(int(time.time()));
    
    self.forcedDongles_ = [];
    self.numUsedDongles_ = 1;
    self.workClasses_ = [];
    
    self.logger_.info("Worker initialized.");

  def lockDongles(self):
    # TODO: subject to deadlock - mitigating with random length sleep
    sleepTime = (self.randomSource_.random()*2); #+1;
    self.logger_.debug("Setting sleep for busy dongles to: "+str(sleepTime));
    self.logger_.debug("Locking forced dongles.");      
    for dongle in self.forcedDongles_:
      d = "";
      while d == "":
        self.logger_.debug("Locking dongle: " + dongle);
        d = self.dongleManager_.lockDongle(dongle);
        if d == "": time.sleep(sleepTime);
      self.dongles_.append(d);
    self.logger_.debug("Locking other dongles.");
    while len(self.dongles_) < self.numUsedDongles_:
      d = "";
      while d == "":
        d = self.dongleManager_.lockDongle();
        if d == "": time.sleep(sleepTime);
      self.logger_.debug("Locking dongle: " + d);
      self.dongles_.append(d);        

  def unlockDongles(self):
    self.logger_.debug("Releasing dongles.");
    for dongle in self.dongles_[:]:
      self.dongleManager_.unlockDongle(dongle);
      self.dongles_.remove(dongle);

  def run(self):
    # Do something
    while True:
      try:
        self.wakeupEvent_.wait();
        self.logger_.debug("Waking up.");
        if self.quitEvent_.isSet():
          self.logger_.debug("Time to leave.");
          break;
        
        ##### LOCK DONGLES #####
        if self.simulation_ == True:
          self.logger_.debug("###### SIMULATION ######");

        if (len(self.workInProgress_) > 0) and (self.simulation_ == False):
          self.lockDongles();
        
        self.logger_.debug("Now to the dirty work.");
        self.doProcessing();
  
        ##### RELEASE DONGLES #####
        self.unlockDongles();

        self.device_ = "";
        self.dongles_ = [];
        self.workInProgress_ = [];
        self.busy_ = False;        
        self.wakeupEvent_.clear();

        if self.quitEvent_.isSet():
          self.logger_.debug("Time to leave.");
          break;
      except:
        exctype, excvalue = sys.exc_info()[:2];
        self.logger_.warning("Caught exception in worker, trying to ignore it: "+str(excvalue));
        self.device_ = "";
        self.dongles_ = [];
        self.workInProgress_ = [];
        self.busy_ = False;        
        self.wakeupEvent_.clear();

    self.logger_.info("HciWorker thread stopping.");
    self.cleanup();

  def cleanup(self):
    # Insert here cleanup to be done right before stopping the thread
    return;
      
  def quitASAP(self):
    self.logger_.debug("Request the worker to quit ASAP.");
    self.quitEvent_.set();
    self.wakeupEvent_.set();

  def signalQueueListener(self, device, work):
    self.logger_.debug("Request the worker to wake up.");
    self.busy_ = True;
    self.device_ = device;
    self.workInProgress_ = work[:];
    self.wakeupEvent_.set();

  def doProcessing(self):
    # Do the actual work using the assigned dongle
    ####
    # wlist = self.workInProgress_[0];
    # for w in self.workInProgress_[1:]:
    #   wlist = wlist + "|" + w; 
    # self.logger_.debug("Working... - Device: " + self.device_ + " - Work : " + wlist);
    ####
    return;

def main():
  # Setting up logging facilities
  logging.basicConfig(level=logging.DEBUG,
                  format='%(asctime)s %(module)s %(levelname)s %(message)s',
                  filename="./hciworker.log",
                  filemode='w');
  console = logging.StreamHandler();
  console.setLevel(logging.DEBUG);
  formatter = logging.Formatter("%(asctime)s %(module)s %(levelname)s %(message)s");
  console.setFormatter(formatter);
  logging.getLogger('').addHandler(console);
  
  # Do some testing
  dongleManager = DongleManager();
  dongleManager.addDongle("hci0");
  dongleManager.addDongle("hci1");
  dongleManager.addDongle("hci2");
  dongleManager.lockDongle("hci2");
  hciWorker = HciWorker(dongleManager);
  hciWorker.forcedDongles_ = ["hci0"];
  hciWorker.numUsedDongles_ = 2;
  hciWorker.start();
  for count in range(1,3):
    hciWorker.signalQueueListener("00:00:00:00:00:0"+str(count), ["work"+str(count), "work"+str(count+1)]);
    time.sleep(2);
  hciWorker.quitASAP();
  hciWorker.join();
  dongleManager.unlockDongle("hci2");

if __name__ == "__main__":
  main();
