#!/usr/bin/python

import BluebagConfig
from DongleManager import DongleManager
from DatabaseManager import DatabaseManager
from threading import Thread, Event
import logging, sys, os, subprocess, signal, traceback
from time import sleep, time

class QueueManager(Thread):

  def __init__(self, workerList, bluebagConfig):
    Thread.__init__(self);
    self.logger_ = logging.getLogger('QueueManager');
    self.bluebagConfig_ = bluebagConfig;
    self.databaseManager_ = None;
    self.pendingDevices_ = [];
    self.deviceQueue_ = {};
    self.timestampQueue_ = [];
    self.assignedDevices_ = [];
    self.simulationList_ = [];
    self.workerList_ = workerList;
    self.quitEvent_ = Event();
    self.wakeupEvent_ = Event();
    self.logger_.info("QueueManager initialized.");

  def run(self):
    self.databaseManager_ = DatabaseManager(self.bluebagConfig_);
    self.databaseManager_.connect();
    while True:
      if self.quitEvent_.isSet():
        self.logger_.debug("Time to leave.");
        break;
      try:
        if len(self.pendingDevices_) == 0: self.wakeupEvent_.clear();
#        if len(self.deviceQueue_) == 0: self.wakeupEvent_.clear();
        self.wakeupEvent_.wait();
        self.logger_.debug("Waking up.");
        if self.quitEvent_.isSet():
          self.logger_.debug("Time to leave.");
          break;
  
        ### Check pending devices (insert/update)
#        for device in self.deviceQueue_.keys():
        for deviceTuple in self.timestampQueue_[:]:
          if self.quitEvent_.isSet(): break;

          device = deviceTuple[1];
          if device not in self.assignedDevices_:
            cursor = self.databaseManager_.execute("SELECT * FROM btdevice WHERE bt_address = '"+device+"' AND bt_task = '"+self.bluebagConfig_.task_+"'");
            res = cursor.fetchall();
            self.databaseManager_.closeCursor();
            workList = [];
            if self.bluebagConfig_.task_ == "survey":
              if (len(res)==0) or ((len(res) != 0) and (res[0][1] == "")):
                self.logger_.info("Trying to fill name field for device "+device);
                workList.append("name");
              if (len(res)==0) or ((len(res) != 0) and res[0][3] == -1):
                self.logger_.info("Trying to fill signal field for device "+device);
                workList.append("lq");
              if (len(res)==0) or ((len(res) != 0) and (res[0][6] == "")):
                self.logger_.info("Trying to fill info field for device "+device);
                workList.append("info");
              if (len(res)==0) or ((len(res) != 0) and (res[0][7] == "")):
                self.logger_.info("Trying to fill sdp field for device "+device);
                workList.append("sdp");
              #if len(res) > 0: self.logger_.debug("***\n"+str(res[0][1])+"***\n"+str(res[0][3])+"***\n"+str(res[0][6])+"***\n"+str(res[0][7])+"***");
            if self.bluebagConfig_.task_ == "stupid":
              if (len(res)==0) or ((len(res) != 0) and (res[0][11] == 0) and (res[0][12] < self.bluebagConfig_.stupidMaxRetry_)):
                self.logger_.info("Trying to fill stupid field for device "+device);
                workList.append("stupid");
            updateDevice = False;
            if len(res) == 0:
              self.logger_.info("New device. Requesting INSERT: "+device);
              updateDevice = False;
            else:
              self.logger_.info("New device. Requesting UPDATE: "+device);
              updateDevice = True;
            self.deviceQueue_[device] = workList;

            assigned = False;
            busyCount = 0;
            while not assigned:
              if self.quitEvent_.isSet(): break;
              for worker in self.workerList_:
                if not worker.busy_:
                  self.logger_.info("Dispatching work list for device "+device+" to worker");
                  self.assignedDevices_.append(device);
                  worker.signalQueueListener(device, self.deviceQueue_[device], updateDevice);
                  assigned = True;
                  busyCount = 0;
                  break;
              if not assigned:
                busyCount+=1;
                if ((busyCount % 5) == 0): self.logger_.info("Workers busy while trying to assign device "+device+". Waiting...");
                sleep(1);
        
        for key, value in self.deviceQueue_.items():
			self.logger_.info("Device in queue: "+key+" - "+str(value));

        for item in self.timestampQueue_:
			self.logger_.info("Timestamp queue: "+str(item[0])+" - "+item[1]);

        if self.quitEvent_.isSet():
          self.logger_.debug("Time to leave.");
          break;

      except:
        self.logger_.fatal("?!??!: " + traceback.format_exc());

    self.cleanup();
    self.logger_.info("QueueManager thread stopping.");

  def signalCompletion(self, device, workList):
    for work in workList:
      self.logger_.debug("Removing \""+work+"\" activity for device "+device+".");
      self.deviceQueue_[device].remove(work);
    if len(self.deviceQueue_[device]) == 0:
      self.logger_.debug("Removing device "+device+" from queue.");
      del self.deviceQueue_[device];
      for deviceTuple in self.timestampQueue_:
        if deviceTuple[1] == device:
          self.timestampQueue_.remove(deviceTuple);
      self.assignedDevices_.remove(device);
      if device in self.simulationList_:
        self.logger_.debug("###### REMOVE SIMULATED DEVICE ######");
        self.simulationList_.remove(device);
    return;
  
  def signalNewDevices(self, newDevices):
    for device in newDevices[:]:
#    for deviceTuple in newDevices[:]:
#        device = deviceTuple[1];
      device = device.upper();
      deviceTuple = (time(), device);
      if (device in self.bluebagConfig_.ignoreDevices_) or (device.lower() in self.bluebagConfig_.ignoreDevices_):
        self.logger_.debug("Ignoring device "+device+" as requested by configuration");
        continue;
      if device not in self.deviceQueue_.keys():
        self.logger_.debug("Adding "+device+" to devices queue.");
        self.deviceQueue_[device] = [];
        self.timestampQueue_.append(deviceTuple);
      else:
        self.logger_.debug("Device "+device+" already in devices queue.");
    self.wakeupEvent_.set();
    return;

  def quitASAP(self):
    self.logger_.debug("Request the queue manager to quit ASAP.");
    # Kill external processes and perform specific cleanup
    # in order to be able to quit ASAP
    self.quitEvent_.set();
    self.wakeupEvent_.set();

  def cleanup(self):
    # Insert here cleanup to be done right before stopping the thread
    self.databaseManager_.disconnect();

def main():
  # Setting up logging facilities
  logging.basicConfig(level=logging.DEBUG,
                  format='%(asctime)s %(module)s %(levelname)s %(message)s',
                  filename="./queuemanager.log",
                  filemode='w');
  console = logging.StreamHandler();
  console.setLevel(logging.INFO);
  formatter = logging.Formatter("%(asctime)s %(module)s %(levelname)s %(message)s");
  console.setFormatter(formatter);
  logging.getLogger('').addHandler(console);
  
  # Do some testing
  dongleManager = DongleManager();
  dongleManager.addDongle("hci0");
  dongleManager.addDongle("hci1");
  queueManager = QueueManager(None);
  from HciDeviceWorker import HciDeviceWorker
  workerList = [];
  hciDeviceWorker = HciDeviceWorker(dongleManager, queueManager, self.bluebagConfig_.taskTag_);
  workerList.append(hciDeviceWorker);
  queueManager.workerList_ = workerList;
  from HciScanWorker import HciScanWorker
  hciScanWorker = HciScanWorker(dongleManager, queueManager);
  try:
    queueManager.start();
    hciDeviceWorker.start();
    hciScanWorker.start();
    sleep(600);
    hciScanWorker.quitASAP();
    hciScanWorker.join();
    queueManager.quitASAP();
    queueManager.join();
    hciDeviceWorker.quitASAP();
    hciDeviceWorker.join();
  except:
    exctype, excvalue = sys.exc_info()[:2];
    logging.getLogger('').fatal("?!??!: " + str(excvalue));
    hciScanWorker.quitASAP();
    hciScanWorker.join();
    queueManager.quitASAP();
    queueManager.join();
    hciDeviceWorker.quitASAP();
    hciDeviceWorker.join();

if __name__ == "__main__":
  main();
