#!/usr/bin/python  

import bluebag
import BluebagConfig
from DongleManager import DongleManager
from DatabaseManager import DatabaseManager
from HciDeviceWorker import HciDeviceWorker
from HciScanWorker import HciScanWorker
from FangWorker import FangWorker
from QueueManager import QueueManager

import logging, sys, time, obexftp, traceback, os, getopt
from threading import Event, Thread, Timer
from random import Random

### GLOBALS ###

os.chdir(os.path.dirname(__file__));

randomSource_ = Random();
randomSource_.jumpahead(int(time.time()));

# Get configuration values
bluebagConfig_ = BluebagConfig.BluebagConfig();

numFangWorkers_ = 0;
fangRangeCenterAddress_ = bluebagConfig_.fangDefaultCenterAddress_;
fangRangeWidth_ = bluebagConfig_.fangDefaultRangeWidth_;

### CLASSES ###

class BluebagStats(Thread):

  def __init__(self, hciScanWorker, startTimestamp, bluebagConfig):
    Thread.__init__(self);
    self.bluebagConfig_ = bluebagConfig;
    self.taskTag_ = "";
    self.quitEvent_ = Event();
    self.databaseManager_ = None;
    self.hciScanWorker_ = hciScanWorker;
    self.elapsedTime_ = 0;
    self.startTime_ = startTimestamp;
    self.numDongles_ = 0;
    self.numFreeDongles_ = 0;
    self.busyFangers_ = 0;
    self.busyScanners_ = 0;
    self.numDevicesInQueue_ = 0;
    self.numStarvingDevices_ = 0;
    self.numFoundDevices_ = 0;
    fangerRanges_ = [];
    self.logger_ = logging.getLogger('BluebagStats');
    self.logger_.info("BluebagStats initialized.");    
  
  def run(self):
    self.databaseManager_ = DatabaseManager(self.bluebagConfig_);
    self.databaseManager_.connect();
    sleepStart = time.time();
    while not self.quitEvent_.isSet():
      now = time.time();
      self.elapsedTime_ = (now - self.startTime_) / 60;
      if (now - sleepStart) < 60: #BluebagConfig.statsInterval_:
        time.sleep(1);
        continue;
      self.generateStats();
      sleepStart = now;
    self.logger_.debug("Cleaning up!");
    self.generateStats();
    self.databaseManager_.disconnect();
    self.databaseManager_ = None;

  def generateStats(self):    
    cursor = self.databaseManager_.execute("SELECT bt_address FROM btdevice WHERE bt_task = '"+self.bluebagConfig_.task_+"' AND bt_tag LIKE '"+self.taskTag_+"%'");
    res = None;
    if cursor:
      res = cursor.fetchall();
      self.databaseManager_.closeCursor();
      self.numFoundDevices_ = len(res);
    
    badGuys = "";
    numStupid = "";
    if self.numFoundDevices_ > 0:
      # Compute the most frequent manufacturer
      mancount = {};
      for addr in res:
        if addr[0][0:8] in mancount.keys():
          mancount[addr[0][0:8]] += 1;
        else:
          mancount[addr[0][0:8]] = 1;
      num = max(mancount.values());
      manid = "";
      for key in mancount.keys():
        if mancount[key] == num: manid = key;
      # Update the fang range
      if manid != bluebag.fangRangeCenterAddress_[0:8]:
        cursor = self.databaseManager_.execute("SELECT bt_address FROM btdevice WHERE bt_address LIKE '"+manid+"%'");
        res = cursor.fetchall();
        self.databaseManager_.closeCursor();
        intresults = []
        for count in range(len(res)): intresults.append(int("0x"+res[count][0].replace(":",""),0));
        intmaxaddr = max(intresults);
        intminaddr = min(intresults);
        intcenter = intmaxaddr + (intmaxaddr - intminaddr)/2;
        strcenter = str(hex(intcenter))[2:-1].zfill(12);
        bluebag.fangRangeCenterAddress_ = strcenter[:2]+":"+strcenter[2:4]+":"+strcenter[4:6]+":"+strcenter[6:8]+":"+strcenter[8:10]+":"+strcenter[10:12];
        self.logger_.debug("New fang range center: "+bluebag.fangRangeCenterAddress_);
        self.logger_.debug("New fang range width: "+str(bluebag.fangRangeWidth_));
        self.fangerRanges_ = splitRange(computeAddressRange(bluebag.fangRangeCenterAddress_, bluebag.fangRangeWidth_), bluebag.numFangWorkers_);
        self.logger_.debug("New fang ranges: "+str(self.fangerRanges_));

    statsText = "Elapsed time: " + str(int(self.elapsedTime_))+"\n"+\
                "Remaining time: " + str(int(self.bluebagConfig_.sessionLength_-self.elapsedTime_))+"\n"+\
                "Num found devices: "+str(self.numFoundDevices_)+"\n"+\
                "Num dongles: "+str(self.numDongles_)+"\n"+\
                "Num free dongles: "+str(self.numFreeDongles_)+"\n"+\
                "Busy fangers: "+str(self.busyFangers_)+"\n"+\
                "Busy scanners: "+str(self.busyScanners_)+"\n"+\
                "Devices in queue: "+str(self.numDevicesInQueue_)+"\n"+\
                "Starving devices: "+str(self.numStarvingDevices_)+"\n"+\
                "Fang center: "+bluebag.fangRangeCenterAddress_+"\n"+\
                "Fang width: "+str(bluebag.fangRangeWidth_)+"\n";
      
    if self.bluebagConfig_.task_ == "survey":
      # Look for problematic devices
      cursor = self.databaseManager_.execute("SELECT bt_address,man_id,bt_name,bt_timesseen FROM btdevice WHERE (bt_name = '' OR bt_info = '' OR bt_servicelist = '' OR bt_signal = '' OR bt_signal = -1) AND (bt_timesseen > 1) AND bt_task = '"+self.bluebagConfig_.task_+"' AND bt_tag LIKE '"+self.taskTag_+"%'");
      res = None;
      if cursor:
        res = cursor.fetchall();
        self.databaseManager_.closeCursor();
        badGuys = "Hostile devices ("+str(len(res))+"):\n";
        if len(res) > 0:
          badGuys += "###\n";
          for row in res[-5:]:
            badGuys += "Address: "+row[0]+"\n";
            badGuys += "ManID: "+row[1]+"\n";
            badGuys += "Name: "+row[2]+"\n";
            badGuys += "Times seen: "+str(row[3])+"\n";
            badGuys += "###\n";
        statsText += badGuys;
        
    if self.bluebagConfig_.task_ == "stupid":
      cursor = self.databaseManager_.execute("SELECT bt_address FROM btdevice WHERE stupid = 1 AND bt_task = '"+self.bluebagConfig_.task_+"' AND bt_tag LIKE '"+self.taskTag_+"%'");
      res = None;
      if cursor:
        res = cursor.fetchall();
        self.databaseManager_.closeCursor();
        numStupid = "Stupid people: "+str(len(res))+"\n";
        statsText += numStupid;
      
    self.logger_.info("Statistics:");
    self.logger_.info(statsText);
    
    stats_fp = open(self.bluebagConfig_.logDir_ + self.bluebagConfig_.task_ + ".status","w");
    stats_fp.write(statsText);
    stats_fp.close();

    #self.logger_.debug("Requesting the scan worker to pause...");
    #self.hciScanWorker_.restart_.clear();
    #while not self.hciScanWorker_.paused_:
    #  self.logger_.debug("Waiting for the scan worker to pause...");
    #  time.sleep(1);
    #for human in BluebagConfig.humanAckAddresses_:
    #  res = "";
    #  cli = None;
    #  cli = obexftp.client(obexftp.BLUETOOTH);
    #  if cli:
    #    res = cli.connect(human,10); # 10, 12 ?
    #    self.logger_.debug("OBEX connect to "+human+": "+str(res));
    #    res = cli.put_data(statsText,"Bluebag-"+time.strftime("%H.%M.%S", time.localtime()));
    #    self.logger_.debug("OBEX put_data to "+human+": "+str(res));
    #    res =  cli.disconnect();
    #    self.logger_.debug("OBEX disconnect to "+human+": "+str(res));
    #    cli.delete;
    #self.logger_.debug("Telling the scan worker to restart...");
    #self.hciScanWorker_.restart_.set();
    
  def setThreadsValues(self,numDongles, numFreeDongles, busyFangers, busyScanners, numDevicesInQueue, numStarvingDevices):
    self.numDongles_ = numDongles;
    self.numFreeDongles_ = numFreeDongles;
    self.busyFangers_ = busyFangers;
    self.busyScanners_ = busyScanners;
    self.numDevicesInQueue_ = numDevicesInQueue;
    self.numStarvingDevices_ = numStarvingDevices;

  def getElapsedTime(self):
    return self.elapsedTime_;

### FUNCTIONS ###

def computeAddressRange(centralAddress, rangeWidth):
  startAddress = "";
  start = 0;
  endAddress = "";
  end = 0;
  address = int("0x"+centralAddress.replace(":",""),0);
  if address < rangeWidth:
    start = 0;
    startAddress = "00:00:00:00:00:00";
  else:
    start = address-(rangeWidth/2);
    start_str = str(hex(start))[2:-1].zfill(12);
    startAddress = start_str[:2]+":"+start_str[2:4]+":"+start_str[4:6]+":"+start_str[6:8]+":"+start_str[8:10]+":"+start_str[10:12];
  if (0xffffffffffff-(rangeWidth/2)) < address:
    end = 0xffffffffffff;
    endAddress = "ff:ff:ff:ff:ff:ff";
  else:
    end = address+(rangeWidth/2);
    end_str = str(hex(end))[2:-1].zfill(12);
    endAddress = end_str[:2]+":"+end_str[2:4]+":"+end_str[4:6]+":"+end_str[6:8]+":"+end_str[8:10]+":"+end_str[10:12];
  
  return startAddress+"-"+endAddress;

def splitRange(addrrange, num):
  if num == 0:
    return [];
  rangeList = [];
  start = int("0x"+addrrange[:17].replace(":",""),0);
  end = int("0x"+addrrange[18:].replace(":",""),0);
  width = (start-end)/num;

  address1 = addrrange[:17];

  addr = start;
  for i in range(num):
    address2 = "";
    if i == (num-1):
      addr = end;
    else:
      addr += width;
    addr_str = str(hex(addr))[2:-1].zfill(12);
    address2 = addr_str[:2]+":"+addr_str[2:4]+":"+addr_str[4:6]+":"+addr_str[6:8]+":"+addr_str[8:10]+":"+addr_str[10:12];
    rangeList.append(address1+"-"+address2);
    address1 = address2;

  return rangeList;

### MAIN ###

def main():
  
  # Parse command line options #
  opts, args = getopt.getopt(sys.argv[1:], "t:", ["task="]);
  for o, a in opts:
    if o in ("-t","--task"):
      bluebagConfig_.task_ = a;
  
  # Setting up logging facilities
  logging.basicConfig(level=bluebagConfig_.logLevel_,
                  format='%(asctime)s %(module)s-%(threadName)s %(levelname)s %(message)s',
                  filename=bluebagConfig_.logDir_+bluebagConfig_.logFilename_,
                  filemode='a');
  console = logging.StreamHandler();
  console.setLevel(bluebagConfig_.consoleLogLevel_);
  formatter = logging.Formatter("%(asctime)s %(module)s-%(threadName)s %(levelname)s %(message)s");
  console.setFormatter(formatter);
  logging.getLogger('').addHandler(console);  

  logging.getLogger('').info("Bluebag starting...");
  
  # HUMAN ACK!
  #for human in BluebagConfig.humanAckAddresses_:
  #  helloText = "Hi, I'm here! Almost ready to start...";
  #  res = "";
  #  cli = None;
  #  cli = obexftp.client(obexftp.BLUETOOTH);
  #  if cli:
  #    res = cli.connect(human,10); # 10, 12 ?
  #    logging.getLogger('').debug("OBEX connect to "+human+": "+str(res));
  #    res = cli.put_data(helloText,"Bluebag-"+time.strftime("%H.%M.%S", time.localtime()));
  #    logging.getLogger('').debug("OBEX put_data to "+human+": "+str(res));
  #    res =  cli.disconnect();
  #    logging.getLogger('').debug("OBEX disconnect to "+human+": "+str(res));
  #    cli.delete;

  # Update the survey tag using the timestamp
  taskTag = bluebagConfig_.taskTag_;
  startTimestamp_ = time.time();
  if bluebagConfig_.tagWithTimestamp_:
    taskTag = taskTag + "-" + time.strftime("%Y%m%d%H%M%S", time.localtime(startTimestamp_));
  
  # Initialize the dongle manager
  dongleManager = DongleManager(bluebagConfig_);

  # Initialize the queue manager
  queueManager = QueueManager(None, bluebagConfig_);

  # Initialize the device workers
  workerList = [];
  #numDeviceWorkers_ = dongleManager.numDongles_ - BluebagConfig.fangMinDongles_ - 1; # 1 dongle reserved for scanning  
  numDeviceWorkers_ = dongleManager.numDongles_; # - 1;
  if bluebagConfig_.simulateDevices_: numDeviceWorkers_ += 2;
  for count in range(0,numDeviceWorkers_):
    hciDeviceWorker = HciDeviceWorker(dongleManager, queueManager, bluebagConfig_, taskTag);
    workerList.append(hciDeviceWorker);
  queueManager.workerList_ = workerList;
  logging.getLogger('').info("Initialized "+str(len(queueManager.workerList_))+" device workers.");

  # Initialize the fang workers
  fangerList_ = [];
  tuple = ();
#  bluebag.numFangWorkers_ = dongleManager.numDongles_ - 1;
  bluebag.numFangWorkers_ = 0;
  fangerRanges_ = splitRange(computeAddressRange(bluebag.fangRangeCenterAddress_, bluebag.fangRangeWidth_), bluebag.numFangWorkers_);
  for count in range(0,bluebag.numFangWorkers_):
    fangWorker = FangWorker(dongleManager, queueManager, bluebagConfig_);
    tuple = (fangWorker, fangerRanges_[count]);
    fangerList_.append(tuple);
  logging.getLogger('').info("Initialized "+str(len(fangerList_))+" fang workers.");

  # Initialize the scan worker
  hciScanWorker = HciScanWorker(dongleManager, queueManager, bluebagConfig_);

  statsThread_ = BluebagStats(hciScanWorker, startTimestamp_, bluebagConfig_);
  statsThread_.taskTag_ = taskTag;
  statsThread_.fangerRanges_ = fangerRanges_;
  statsThread_.start();
  
  try:
    hciScanWorker.start();
    time.sleep(1);
    dongleManager.start();
    time.sleep(1);
    queueManager.start();
    time.sleep(1);
    for count in range(0,numDeviceWorkers_):
      workerList[count].start();
      time.sleep(1);
    for count in range(0,bluebag.numFangWorkers_):
      fangerList_[count][0].start();
      time.sleep(1);
    
    surveyTimeElapsed = False;
    simulCount = 0;
    while not surveyTimeElapsed:
      if (bluebagConfig_.sessionLength_ > 0) and (statsThread_.elapsedTime_ > bluebagConfig_.sessionLength_): surveyTimeElapsed = True;

      # Simulate devices
      simulCount += 1;
      if bluebagConfig_.simulateDevices_ and simulCount == 2:
        simulCount = 0;
        simulAddresses = [];
        for i in range(randomSource_.randint(1,3)):
          simulAddress = bluebagConfig_.simulationBaseAddress_;
          simulAddress = simulAddress.replace("V",str(hex(randomSource_.randint(0,15)))[2:]);
          simulAddress = simulAddress.replace("W",str(hex(randomSource_.randint(0,15)))[2:]);
          simulAddress = simulAddress.replace("X",str(hex(randomSource_.randint(0,15)))[2:]);
          simulAddress = simulAddress.replace("Y",str(hex(randomSource_.randint(0,15)))[2:]);
          simulAddress = simulAddress.replace("Z",str(hex(randomSource_.randint(0,15)))[2:]);
          simulAddresses.append(simulAddress.upper());
          queueManager.simulationList_.append(simulAddress.upper());
        queueManager.signalNewDevices(simulAddresses);
      
      # Analyze detected devices and select new brute-force range            
      # Start/stop fang workers as needed
      numFreeDongles = 0;
      for key, value in dongleManager.busyDongles_.items():
        if value == False: numFreeDongles += 1;

      numBusyFangers = 0;
      for fanger in fangerList_:
        if fanger[0].busy_: numBusyFangers += 1;
      numBusyScanners = len(queueManager.assignedDevices_);
      starvingDevices = numBusyFangers + numBusyScanners - dongleManager.numDongles_ + 1; # 1 dongle is busy scanning
      statsThread_.setThreadsValues(dongleManager.numDongles_, numFreeDongles, numBusyFangers, numBusyScanners, len(queueManager.deviceQueue_), starvingDevices);

      logging.getLogger('').info("Free dongle: "+str(numFreeDongles)+" - Busy fangers: "+str(numBusyFangers)+" - Busy scanners: "+str(numBusyScanners)+" - Devices in queue: "+str(len(queueManager.deviceQueue_))+" - Starving devices: "+str(starvingDevices));

      for count in range(0,bluebag.numFangWorkers_):
        fangWorker = fangerList_[count][0];
        tuple = (fangWorker, statsThread_.fangerRanges_[count]);
        fangerList_[count] = tuple;

      if numFreeDongles > 0:
        # Restart fang workers
        numStart = min(numFreeDongles, bluebag.numFangWorkers_)
        logging.getLogger('').debug("Starting "+str(numStart)+" fanger(s)");
        for i in range(numStart):
          for fanger in fangerList_:
            if not fanger[0].busy_:
              fanger[0].signalQueueListener(fanger[1], ["fang"]);
      else:
        # Stop fang workers
        count = min(starvingDevices, len(fangerList_));
        logging.getLogger('').debug("Stopping "+str(count)+" fanger(s)");
        for i in range(count):
          fangerList_[i][0].stopFang();
          fangerList_[i] = (fangerList_[i][0], fangerList_[i][0].remainingRange_);
      
      # Take a deep breath...
      time.sleep(3);
      
#  finally:
  except:
    logging.getLogger('').fatal("?!??!: "+traceback.format_exc());

  logging.getLogger('').info("Quitting...");
  hciScanWorker.quitASAP();
  hciScanWorker.join();
  hciScanWorker = None;
  queueManager.quitASAP();
  queueManager.join();
  queueManager = None;

  for count in range(0,numDeviceWorkers_):
    workerList[count].quitASAP();
  for count in range(0,numDeviceWorkers_):
    workerList[count].join();
  workerList = [];
  for count in range(0,bluebag.numFangWorkers_):
    fangerList_[count][0].quitASAP();
  for count in range(0,bluebag.numFangWorkers_):
    fangerList_[count][0].join();
  fangerList = [];

  dongleManager.quitASAP();
  dongleManager.join();
  dongleManager = None;

  statsThread_.quitEvent_.set();
  statsThread_.join();
  
  logging.getLogger('').info("Cleaned up. Bye!");

  os.system("sync");

  if bluebagConfig_.shutdown_ == True:
      logging.getLogger('').info("Survey ended. Shutting down...");
      # HUMAN ACK!
      #for human in BluebagConfig.humanAckAddresses_:
      #  byeText = "Guys, I've cleaned up everything.\n"+\
      #            "I also did a sync to flush my beloved buffers.\n"+\
      #            "Now I need to rest a little bit.\n"+\
      #            "'nite!";
      #  res = "";
      #  cli = None;
      #  cli = obexftp.client(obexftp.BLUETOOTH);
      #  if cli:
      #    res = cli.connect(human,10); # 10, 12 ?
      #    logging.getLogger('').debug("OBEX connect to "+human+": "+str(res));
      #    res = cli.put_data(byeText,"Bluebag-"+time.strftime("%H.%M.%S", time.localtime()));
      #    logging.getLogger('').debug("OBEX put_data to "+human+": "+str(res));
      #    res =  cli.disconnect();
      #    logging.getLogger('').debug("OBEX disconnect to "+human+": "+str(res));
      #    cli.delete;

      os.system("shutdown -h +15");

if __name__ == "__main__":
  main();
  
